// src/main.rs
#![no_std]
#![no_main]
#![feature(naked_functions)]

use core::panic::PanicInfo;

mod heap;
mod uart;

static INIT_LOCK: spin::Once<()> = spin::Once::new();

core::arch::global_asm!(include_str!("entry.s"));

/// After some initialization in asm/entry.s, the kernel will jump here and
/// each hart will have its own setup sequence.
#[no_mangle]
unsafe extern "C" fn start() {
    use core::arch::asm;
    use riscv::register::{mepc, mstatus, satp, sie, sstatus};

    // Set previous privilege for all harts to M-mode, set previous interrupt
    // enable, and set floating-point unit to initial state
    mstatus::set_mpp(mstatus::MPP::Supervisor);
    mstatus::set_mpie();
    mstatus::set_fs(sstatus::FS::Initial);

    // Disable paging
    satp::write(0);

    // Delegate all traps to S-mode, using inline assembly since we have not
    // provided a wrapper for it yet
    asm!("li t0, 0xffff", "csrw medeleg, t0", "csrw mideleg, t0",);

    // Enable S-mode external, timer, and software interrupts
    sie::set_sext();
    sie::set_stimer();
    sie::set_ssoft();

    // TODO configure PMP
    // TODO timer init

    // Trying to figure out why I can't use mret here. Might need
    // asm!(
    //     "la t1, 1f",
    //     "csrw mepc, t1",
    //     "la ra, 1f",
    //     "mret",
    // "1:",
    // );

    main();
}

fn main() {
    use riscv::interrupt;
    use riscv::register::{mstatus, sstatus};

    if hartid() == 0 {
        INIT_LOCK.call_once(|| {
            // Disable machine interrupts while initializing
            interrupt::machine::disable();
            console_init();
            // TODO Write boot message

            // TODO Set up paging
            // TODO Set up processes

            // TODO Set up trap vectors
            // TODO Set up PLIC

            kinit();
            unsafe {
                mstatus::set_mpp(mstatus::MPP::User);
                mstatus::set_mpie();
                mstatus::set_spie();
                mstatus::set_fs(sstatus::FS::Initial);
            }

            riscv::asm::fence(); // Emit a fence just in case
        });

        // poll console for input and print characters back
        loop {
            let c = uart::CONSOLE.lock().as_mut().and_then(uart::Device::get);
            if let Some(c) = c { print!("{}", c as char); }
        }
    } else {
        INIT_LOCK.wait();
        riscv::asm::fence(); // Emit a fence just in case
        kinit_hart();
        println!("Hello from hart {}!", hartid());
    }
}

fn console_init() {
    // Initialize heap
    unsafe { heap::init(); };

    // Set up UART and print to console
    uart::init_console(0x1000_0000);
    println!("Hello from hart {}!", hartid());
}

fn kinit() {}

fn kinit_hart() {}

#[inline]
fn hartid() -> usize {
    use core::arch::asm;
    let id: usize;
    unsafe {
        asm!(
            "mv {id}, tp",
            id = out(reg) id,
        );
    }
    id
}

#[no_mangle]
extern "C" fn eh_personality() {}

#[panic_handler]
fn panic(info: &PanicInfo) -> ! {
    // print panic info and abort
    println!("{}", info);
    abort();
}

#[no_mangle]
extern "C" fn abort() -> ! {
    // loop wait for interrupt
    loop { riscv::asm::wfi(); }
}

// TODO unit testing